@@ -3,22 +3,42 @@ import requests
 import sys
 from base64 import b64encode
 from datetime import datetime
+from enum import StrEnum
 from logging import Logger
 from pypomes_core import TZ_LOCAL, exc_format
 from threading import Lock
 from typing import Any, Final
 
+
+class ProviderParam(StrEnum):
+    """
+    Parameters for configuring a *JWT* token provider.
+    """
+    URL = "url"
+    USER = "user"
+    PWD = "pwd"
+    CUSTOM_AUTH = "custom-auth"
+    HEADER_DATA = "headers-data"
+    BODY_DATA = "body-data"
+    ACCESS_TOKEN = "access-token"
+    ACCESS_EXPIRATION = "access-expiration"
+    REFRESH_TOKEN = "refresh-token"
+    REFRESH_EXPIRATION = "refresh-expiration"
+
+
 # structure:
 # {
 #    <provider-id>: {
 #      "url": <strl>,
 #      "user": <str>,
 #      "pwd": <str>,
-#      "basic-auth": <bool>,
+#      "custom-auth": <bool>,
 #      "headers-data": <dict[str, str]>,
 #      "body-data": <dict[str, str],
 #      "access-token": <str>,
-#      "access-expiration": <timestamp>
+#      "access-expiration": <timestamp>,
+#      "refresh-token": <str>,
+#      "refresh-expiration": <timestamp>
 #    }
 # }
 _provider_registry: Final[dict[str, dict[str, Any]]] = {}
@@ -58,16 +78,16 @@ def provider_register(provider_id: str,
 
     with _provider_lock:
         _provider_registry[provider_id] = {
-            "url": auth_url,
-            "user": auth_user,
-            "pwd": auth_pwd,
-            "custom-auth": custom_auth,
-            "headers-data": headers_data,
-            "body-data": body_data,
-            "access-token": None,
-            "access-expiration": 0,
-            "refresh-token": None,
-            "refresh-expiration": 0
+            ProviderParam.URL: auth_url,
+            ProviderParam.USER: auth_user,
+            ProviderParam.PWD: auth_pwd,
+            ProviderParam.CUSTOM_AUTH: custom_auth,
+            ProviderParam.HEADER_DATA: headers_data,
+            ProviderParam.BODY_DATA: body_data,
+            ProviderParam.ACCESS_TOKEN: None,
+            ProviderParam.ACCESS_EXPIRATION: 0,
+            ProviderParam.REFRESH_TOKEN: None,
+            ProviderParam.REFRESH_EXPIRATION: 0
         }
 
 
@@ -91,19 +111,19 @@ def provider_get_token(provider_id: str,
         provider: dict[str, Any] = _provider_registry.get(provider_id)
         if provider:
             now: float = datetime.now(tz=TZ_LOCAL).timestamp()
-            if now > provider.get("access-expiration"):
-                user: str = provider.get("user")
-                pwd: str = provider.get("pwd")
-                headers_data: dict[str, str] = provider.get("headers-data") or {}
-                body_data: dict[str, str] = provider.get("body-data") or {}
-                custom_auth: tuple[str, str] = provider.get("custom-auth")
+            if now > provider.get(ProviderParam.ACCESS_EXPIRATION):
+                user: str = provider.get(ProviderParam.USER)
+                pwd: str = provider.get(ProviderParam.PWD)
+                headers_data: dict[str, str] = provider.get(ProviderParam.HEADER_DATA) or {}
+                body_data: dict[str, str] = provider.get(ProviderParam.BODY_DATA) or {}
+                custom_auth: tuple[str, str] = provider.get(ProviderParam.CUSTOM_AUTH)
                 if custom_auth:
                     body_data[custom_auth[0]] = user
                     body_data[custom_auth[1]] = pwd
                 else:
                     enc_bytes: bytes = b64encode(f"{user}:{pwd}".encode())
                     headers_data["Authorization"] = f"Basic {enc_bytes.decode()}"
-                url: str = provider.get("url")
+                url: str = provider.get(ProviderParam.URL)
                 if logger:
                     logger.debug(msg=f"POST {url}, {json.dumps(obj=body_data,
                                                                ensure_ascii=False)}")
@@ -130,14 +150,14 @@ def provider_get_token(provider_id: str,
                         if logger:
                             logger.debug(msg=f"POST success, status {response.status_code}")
                         reply: dict[str, Any] = response.json()
-                        provider["access-token"] = reply.get("access_token")
-                        provider["access-expiration"] = now + int(reply.get("expires_in"))
-                        if reply.get("refresh_token"):
-                            provider["refresh-token"] = reply["refresh_token"]
+                        provider[ProviderParam.ACCESS_TOKEN] = reply.get("access_token")
+                        provider[ProviderParam.ACCESS_EXPIRATION] = now + int(reply.get("expires_in"))
+                        if reply.get(ProviderParam.REFRESH_TOKEN):
+                            provider[ProviderParam.REFRESH_TOKEN] = reply["refresh_token"]
                             if reply.get("refresh_expires_in"):
-                                provider["refresh-expiration"] = now + int(reply.get("refresh_expires_in"))
+                                provider[ProviderParam.REFRESH_EXPIRATION] = now + int(reply.get("refresh_expires_in"))
                             else:
-                                provider["refresh-expiration"] = sys.maxsize
+                                provider[ProviderParam.REFRESH_EXPIRATION] = sys.maxsize
                         if logger:
                             logger.debug(msg=f"POST {url}: status {response.status_code}")
                 except Exception as e:
@@ -154,7 +174,7 @@ def provider_get_token(provider_id: str,
         if logger:
             logger.error(msg=err_msg)
     else:
-        result = provider.get("access-token")
+        result = provider.get(ProviderParam.ACCESS_TOKEN)
 
     return result
 
